"""ASN.1 constraints checker.

"""

import string
from copy import copy

from . import ConstraintsError
from . import compiler
from . import format_or
from .permitted_alphabet import NUMERIC_STRING
from .permitted_alphabet import PRINTABLE_STRING
from .permitted_alphabet import IA5_STRING
from .permitted_alphabet import BMP_STRING
from .permitted_alphabet import VISIBLE_STRING


STRING_TYPES = [
    'OBJECT IDENTIFIER',
    'TeletexString',
    'NumericString',
    'PrintableString',
    'IA5String',
    'VisibleString',
    'GeneralString',
    'UTF8String',
    'BMPString',
    'GraphicString',
    'UniversalString',
    'ObjectDescriptor'
]

TIME_TYPES = [
    'UTCTime',
    'GeneralizedTime',
    'DATE',
    'TIME-OF-DAY',
    'DATE-TIME'
]


class Type(object):

    def __init__(self, name):
        self.name = name
        self.minimum = 'MIN'
        self.maximum = 'MAX'

    def set_range(self, minimum, maximum, has_extension_marker):
        if has_extension_marker:
            return

        if minimum is None:
            minimum = 'MIN'

        if maximum is None:
            maximum = 'MAX'

        self.minimum = minimum
        self.maximum = maximum

    def set_size_range(self, minimum, maximum, has_extension_marker):
        self.set_range(minimum, maximum, has_extension_marker)

    def set_restricted_to_range(self,
                                minimum,
                                maximum,
                                has_extension_marker):
        self.set_range(minimum, maximum, has_extension_marker)

    def is_in_range(self, value):
        minimum_ok = (not self.has_lower_bound()) or (value >= self.minimum)
        maximum_ok = (not self.has_upper_bound()) or (value <= self.maximum)

        return minimum_ok and maximum_ok

    def set_default(self, value):
        pass

    def has_lower_bound(self):
        return self.minimum != 'MIN'

    def has_upper_bound(self):
        return self.maximum != 'MAX'

    def is_bound(self):
        return self.has_lower_bound() and self.has_upper_bound()

    def encode(self, data):
        raise NotImplementedError('To be implemented by subclasses.')


class String(Type):

    PERMITTED_ALPHABET = None

    def __init__(self,
                 name,
                 permitted_alphabet,
                 minimum,
                 maximum,
                 has_extension_marker):
        super(String, self).__init__(name)
        self.set_size_range(minimum, maximum, has_extension_marker)

        if permitted_alphabet is None:
            permitted_alphabet = self.PERMITTED_ALPHABET

        self.permitted_alphabet = permitted_alphabet

    def encode(self, data):
        length = len(data)

        if not self.is_in_range(length):
            raise ConstraintsError(
                'Expected between {} and {} characters, but got {}.'.format(
                    self.minimum,
                    self.maximum,
                    length))

        if self.permitted_alphabet is None:
            return

        for character in data:
            if character not in self.permitted_alphabet:
                raise ConstraintsError(
                    "Expected a character in '{}', but got '{}' (0x{:02x}).".format(
                        ''.join([c if c in string.printable[:-5] else '.'
                                 for c in self.permitted_alphabet]),
                        character if character in string.printable else '.',
                        ord(character)))


class Boolean(Type):

    def encode(self, data):
        pass


class Integer(Type):

    def __init__(self, name):
        super(Integer, self).__init__(name)

    def encode(self, data):
        if not self.is_in_range(data):
            raise ConstraintsError(
                'Expected an integer between {} and {}, but got {}.'.format(
                    self.minimum,
                    self.maximum,
                    data))


class Float(Type):

    def encode(self, data):
        pass


class Null(Type):

    def encode(self, data):
        pass


class BitString(Type):

    def __init__(self, name, minimum, maximum, has_extension_marker):
        super(BitString, self).__init__(name)
        self.set_size_range(minimum, maximum, has_extension_marker)

    def encode(self, data):
        number_of_bits = data[1]

        if not self.is_in_range(number_of_bits):
            raise ConstraintsError(
                'Expected between {} and {} bits, but got {}.'.format(
                    self.minimum,
                    self.maximum,
                    number_of_bits))


class Enumerated(Type):

    def encode(self, data):
        pass


class Bytes(Type):

    def __init__(self, name, minimum, maximum, has_extension_marker):
        super(Bytes, self).__init__(name)
        self.set_size_range(minimum, maximum, has_extension_marker)

    def encode(self, data):
        length = len(data)

        if not self.is_in_range(length):
            raise ConstraintsError(
                'Expected between {} and {} bytes, but got {}.'.format(
                    self.minimum,
                    self.maximum,
                    length))


class Dict(Type):

    def __init__(self, name, members):
        super(Dict, self).__init__(name)
        self.members = members

    def encode(self, data):
        for member in self.members:
            name = member.name

            if name in data:
                try:
                    member.encode(data[name])
                except ConstraintsError as e:
                    e.location.append(member.name)
                    raise


class List(Type):

    def __init__(self, name, element_type, minimum, maximum, has_extension_marker):
        super(List, self).__init__(name)
        self.element_type = element_type
        self.set_size_range(minimum, maximum, has_extension_marker)

    def encode(self, data):
        length = len(data)

        if not self.is_in_range(length):
            raise ConstraintsError(
                'Expected a list of between {} and {} elements, but got {}.'.format(
                    self.minimum,
                    self.maximum,
                    length))

        for entry in data:
            self.element_type.encode(entry)


class Choice(Type):

    def __init__(self, name, members, has_extension_marker):
        super(Choice, self).__init__(name)
        self.members = members
        self.name_to_member = {member.name: member for member in self.members}
        self.has_extension_marker = has_extension_marker

    def format_names(self):
        return format_or(sorted(self.name_to_member))

    def encode(self, data):
        value = data[0]

        if value in self.name_to_member:
            member = self.name_to_member[value]
        elif self.has_extension_marker:
            return
        else:
            raise ConstraintsError(
                "Expected choice {}, but got '{}'.".format(
                    self.format_names(),
                    value))

        try:
            member.encode(data[1])
        except ConstraintsError as e:
            e.location.append(member.name)
            raise


class NumericString(String):

    PERMITTED_ALPHABET = NUMERIC_STRING


class PrintableString(String):

    PERMITTED_ALPHABET = PRINTABLE_STRING


class IA5String(String):

    PERMITTED_ALPHABET = IA5_STRING


class BMPString(String):

    PERMITTED_ALPHABET = BMP_STRING


class VisibleString(String):

    PERMITTED_ALPHABET = VISIBLE_STRING


class Time(Type):

    def encode(self, data):
        pass


class Skip(Type):

    def encode(self, data):
        pass


class Recursive(Type, compiler.Recursive):

    def __init__(self, name, type_name, module_name):
        super(Recursive, self).__init__(name)
        self.type_name = type_name
        self.module_name = module_name
        self.inner = None

    def set_inner_type(self, inner):
        self.inner = copy(inner)

    def encode(self, data):
        self.inner.encode(data)


class CompiledType(compiler.CompiledType):

    def encode(self, data):
        self._type.encode(data)


class Compiler(compiler.Compiler):

    def process_type(self, type_name, type_descriptor, module_name):
        compiled_type = self.compile_type(type_name,
                                          type_descriptor,
                                          module_name)

        return CompiledType(compiled_type)

    def compile_type(self, name, type_descriptor, module_name):
        module_name = type_descriptor.get('module-name', module_name)
        type_name = type_descriptor['type']

        if type_name in ['SEQUENCE', 'SET']:
            members, _ = self.compile_members(type_descriptor['members'],
                                              module_name)
            compiled = Dict(name, members)
        elif type_name in ['SEQUENCE OF', 'SET OF']:
            element_type = self.compile_type('',
                                             type_descriptor['element'],
                                             module_name)
            compiled = List(name,
                            element_type,
                            *self.get_size_range(
                                type_descriptor,
                                module_name))
        elif type_name == 'CHOICE':
            compiled = Choice(name,
                              *self.compile_members(type_descriptor['members'],
                                                    module_name))
        elif type_name == 'INTEGER':
            compiled = Integer(name)
        elif type_name == 'REAL':
            compiled = Float(name)
        elif type_name == 'BOOLEAN':
            compiled = Boolean(name)
        elif type_name == 'OCTET STRING':
            compiled = Bytes(name,
                             *self.get_size_range(type_descriptor,
                                                  module_name))
        elif type_name == 'ENUMERATED':
            compiled = Enumerated(name)
        elif type_name in TIME_TYPES:
            compiled = Time(name)
        elif type_name == 'BIT STRING':
            compiled = BitString(name,
                                 *self.get_size_range(type_descriptor,
                                                      module_name))
        elif type_name == 'NumericString':
            compiled = NumericString(name,
                                     self.get_permitted_alphabet(type_descriptor),
                                     *self.get_size_range(type_descriptor,
                                                          module_name))
        elif type_name == 'PrintableString':
            compiled = PrintableString(name,
                                       self.get_permitted_alphabet(type_descriptor),
                                       *self.get_size_range(type_descriptor,
                                                            module_name))
        elif type_name == 'IA5String':
            compiled = IA5String(name,
                                 self.get_permitted_alphabet(type_descriptor),
                                 *self.get_size_range(type_descriptor,
                                                      module_name))
        elif type_name == 'BMPString':
            compiled = BMPString(name,
                                 self.get_permitted_alphabet(type_descriptor),
                                 *self.get_size_range(type_descriptor,
                                                      module_name))
        elif type_name == 'VisibleString':
            compiled = VisibleString(name,
                                     self.get_permitted_alphabet(type_descriptor),
                                     *self.get_size_range(type_descriptor,
                                                          module_name))
        elif type_name in STRING_TYPES:
            compiled = String(name,
                              self.get_permitted_alphabet(type_descriptor),
                              *self.get_size_range(type_descriptor,
                                                   module_name))
        elif type_name in ['ANY', 'ANY DEFINED BY', 'OpenType']:
            compiled = Skip(name)
        elif type_name == 'NULL':
            compiled = Null(name)
        elif type_name == 'EXTERNAL':
            members, _ = self.compile_members(
                self.external_type_descriptor()['members'],
                module_name)
            compiled = Dict(name, members)
        else:
            if type_name in self.types_backtrace:
                compiled = Recursive(name,
                                     type_name,
                                     module_name)
                self.recursive_types.append(compiled)
            else:
                compiled = self.compile_user_type(name,
                                                  type_name,
                                                  module_name)

        if 'restricted-to' in type_descriptor:
            compiled = self.set_compiled_restricted_to(compiled,
                                                       type_descriptor,
                                                       module_name)

        return compiled

    def get_permitted_alphabet(self, type_descriptor):
        def char_range(begin, end):
            return ''.join([chr(char)
                            for char in range(ord(begin), ord(end) + 1)])

        if 'from' not in type_descriptor:
            return

        permitted_alphabet = type_descriptor['from']
        value = ''

        for item in permitted_alphabet:
            if isinstance(item, tuple):
                value += char_range(item[0], item[1])
            else:
                value += item

        return ''.join(sorted(value))


def compile_dict(specification, numeric_enums=False):
    return Compiler(specification, numeric_enums).process()



"""Functions required by the UPER C code generator

"""

from .utils import ENCODER_ABORT
from .utils import DECODER_ABORT

ENCODER_INIT = '''\
static void encoder_init(struct encoder_t *self_p,
                         uint8_t *buf_p,
                         size_t size)
{
    self_p->buf_p = buf_p;
    self_p->size = (8 * (ssize_t)size);
    self_p->pos = 0;
}\
'''

ENCODER_GET_RESULT = '''
static ssize_t encoder_get_result(const struct encoder_t *self_p)
{
    if (self_p->size >= 0) {
        return ((self_p->pos + 7) / 8);
    } else {
        return (self_p->pos);
    }
}\
'''

ENCODER_ALLOC = '''
static ssize_t encoder_alloc(struct encoder_t *self_p,
                             size_t size)
{
    ssize_t pos;

    if ((self_p->pos + (ssize_t)size) <= self_p->size) {
        pos = self_p->pos;
        self_p->pos += (ssize_t)size;
    } else {
        pos = -ENOMEM;
        encoder_abort(self_p, ENOMEM);
    }

    return (pos);
}\
'''

ENCODER_APPEND_BIT = '''
static void encoder_append_bit(struct encoder_t *self_p,
                               int value)
{
    ssize_t pos;

    pos = encoder_alloc(self_p, 1);

    if (pos < 0) {
        return;
    }

    if ((pos % 8) == 0) {
        self_p->buf_p[pos / 8] = 0;
    }

    self_p->buf_p[pos / 8] |= (uint8_t)(value << (7 - (pos % 8)));
}\
'''

ENCODER_APPEND_BYTES = '''
static void encoder_append_bytes(struct encoder_t *self_p,
                                 const uint8_t *buf_p,
                                 size_t size)
{
    size_t i;
    ssize_t pos;
    size_t byte_pos;
    size_t pos_in_byte;

    pos = encoder_alloc(self_p, 8u * size);

    if (pos < 0) {
        return;
    }

    byte_pos = ((size_t)pos / 8u);
    pos_in_byte = ((size_t)pos % 8u);

    if (pos_in_byte == 0u) {
        (void)memcpy(&self_p->buf_p[byte_pos], buf_p, size);
    } else {
        for (i = 0; i < size; i++) {
            self_p->buf_p[byte_pos + i] |= (buf_p[i] >> pos_in_byte);
            self_p->buf_p[byte_pos + i + 1] = (buf_p[i] << (8u - pos_in_byte));
        }
    }
}\
'''

ENCODER_APPEND_UINT8 = '''
static void encoder_append_uint8(struct encoder_t *self_p,
                                 uint8_t value)
{
    uint8_t buf[1];

    buf[0] = (uint8_t)value;

    encoder_append_bytes(self_p, &buf[0], sizeof(buf));
}\
'''

ENCODER_APPEND_UINT16 = '''
static void encoder_append_uint16(struct encoder_t *self_p,
                                  uint16_t value)
{
    uint8_t buf[2];

    buf[0] = (uint8_t)(value >> 8);
    buf[1] = (uint8_t)value;

    encoder_append_bytes(self_p, &buf[0], sizeof(buf));
}\
'''

ENCODER_APPEND_UINT32 = '''
static void encoder_append_uint32(struct encoder_t *self_p,
                                  uint32_t value)
{
    uint8_t buf[4];

    buf[0] = (uint8_t)(value >> 24);
    buf[1] = (uint8_t)(value >> 16);
    buf[2] = (uint8_t)(value >> 8);
    buf[3] = (uint8_t)value;

    encoder_append_bytes(self_p, &buf[0], sizeof(buf));
}\
'''

ENCODER_APPEND_UINT64 = '''
static void encoder_append_uint64(struct encoder_t *self_p,
                                  uint64_t value)
{
    uint8_t buf[8];

    buf[0] = (uint8_t)(value >> 56);
    buf[1] = (uint8_t)(value >> 48);
    buf[2] = (uint8_t)(value >> 40);
    buf[3] = (uint8_t)(value >> 32);
    buf[4] = (uint8_t)(value >> 24);
    buf[5] = (uint8_t)(value >> 16);
    buf[6] = (uint8_t)(value >> 8);
    buf[7] = (uint8_t)value;

    encoder_append_bytes(self_p, &buf[0], sizeof(buf));
}\
'''

ENCODER_APPEND_INT8 = '''
static void encoder_append_int8(struct encoder_t *self_p,
                                int8_t value)
{
    encoder_append_uint8(self_p, (uint8_t)value + 128);
}\
'''

ENCODER_APPEND_INT16 = '''
static void encoder_append_int16(struct encoder_t *self_p,
                                 int16_t value)
{
    encoder_append_uint16(self_p, (uint16_t)value + 32768);
}\
'''

ENCODER_APPEND_INT32 = '''
static void encoder_append_int32(struct encoder_t *self_p,
                                 int32_t value)
{
    encoder_append_uint32(self_p, (uint32_t)value + 2147483648);
}\
'''

ENCODER_APPEND_INT64 = '''
static void encoder_append_int64(struct encoder_t *self_p,
                                 int64_t value)
{
    uint64_t u64_value;

    u64_value = (uint64_t)value;
    u64_value += 9223372036854775808ull;

    encoder_append_uint64(self_p, u64_value);
}\
'''

ENCODER_APPEND_BOOL = '''
static void encoder_append_bool(struct encoder_t *self_p, bool value)
{
    encoder_append_bit(self_p, value ? 1 : 0);
}\
'''

ENCODER_APPEND_NON_NEGATIVE_BINARY_INTEGER = '''
static void encoder_append_non_negative_binary_integer(struct encoder_t *self_p,
                                                       uint64_t value,
                                                       size_t size)
{
    size_t i;

    for (i = 0; i < size; i++) {
        encoder_append_bit(self_p, (value >> (size - i - 1)) & 1);
    }
}\
'''

DECODER_INIT = '''
static void decoder_init(struct decoder_t *self_p,
                         const uint8_t *buf_p,
                         size_t size)
{
    self_p->buf_p = buf_p;
    self_p->size = (8 * (ssize_t)size);
    self_p->pos = 0;
}\
'''

DECODER_GET_RESULT = '''
static ssize_t decoder_get_result(const struct decoder_t *self_p)
{
    if (self_p->size >= 0) {
        return ((self_p->pos + 7) / 8);
    } else {
        return (self_p->pos);
    }
}\
'''

DECODER_FREE = '''
static ssize_t decoder_free(struct decoder_t *self_p,
                            size_t size)
{
    ssize_t pos;

    if ((self_p->pos + (ssize_t)size) <= self_p->size) {
        pos = self_p->pos;
        self_p->pos += (ssize_t)size;
    } else {
        pos = -EOUTOFDATA;
        decoder_abort(self_p, EOUTOFDATA);
    }

    return (pos);
}\
'''

DECODER_READ_BIT = '''
static int decoder_read_bit(struct decoder_t *self_p)
{
    ssize_t pos;
    int value;

    pos = decoder_free(self_p, 1);

    if (pos >= 0) {
        value = ((self_p->buf_p[pos / 8] >> (7 - (pos % 8))) & 1);
    } else {
        value = 0;
    }

    return (value);
}\
'''

DECODER_READ_BYTES = '''
static void decoder_read_bytes(struct decoder_t *self_p,
                               uint8_t *buf_p,
                               size_t size)
{
    size_t i;
    ssize_t pos;
    size_t byte_pos;
    size_t pos_in_byte;

    pos = decoder_free(self_p, 8u * size);

    if (pos < 0) {
        return;
    }

    byte_pos = ((size_t)pos / 8u);
    pos_in_byte = ((size_t)pos % 8u);

    if (pos_in_byte == 0) {
        (void)memcpy(buf_p, &self_p->buf_p[byte_pos], size);
    } else {
        for (i = 0; i < size; i++) {
            buf_p[i] = (self_p->buf_p[byte_pos + i] << pos_in_byte);
            buf_p[i] |= (self_p->buf_p[byte_pos + i + 1] >> (8u - pos_in_byte));
        }
    }
}\
'''

DECODER_READ_UINT8 = '''
static uint8_t decoder_read_uint8(struct decoder_t *self_p)
{
    uint8_t value;

    decoder_read_bytes(self_p, &value, sizeof(value));

    return (value);
}\
'''

DECODER_READ_UINT16 = '''
static uint16_t decoder_read_uint16(struct decoder_t *self_p)
{
    uint8_t buf[2];

    decoder_read_bytes(self_p, &buf[0], sizeof(buf));

    return (((uint16_t)buf[0] << 8) | (uint16_t)buf[1]);
}\
'''

DECODER_READ_UINT32 = '''
static uint32_t decoder_read_uint32(struct decoder_t *self_p)
{
    uint8_t buf[4];

    decoder_read_bytes(self_p, &buf[0], sizeof(buf));

    return (((uint32_t)buf[0] << 24)
            | ((uint32_t)buf[1] << 16)
            | ((uint32_t)buf[2] << 8)
            | (uint32_t)buf[3]);
}\
'''

DECODER_READ_UINT64 = '''
static uint64_t decoder_read_uint64(struct decoder_t *self_p)
{
    uint8_t buf[8];

    decoder_read_bytes(self_p, &buf[0], sizeof(buf));

    return (((uint64_t)buf[0] << 56)
            | ((uint64_t)buf[1] << 48)
            | ((uint64_t)buf[2] << 40)
            | ((uint64_t)buf[3] << 32)
            | ((uint64_t)buf[4] << 24)
            | ((uint64_t)buf[5] << 16)
            | ((uint64_t)buf[6] << 8)
            | (uint64_t)buf[7]);
}\
'''

DECODER_READ_INT8 = '''
static int8_t decoder_read_int8(struct decoder_t *self_p)
{
    int8_t value;

    value = (int8_t)decoder_read_uint8(self_p);
    value -= 128;

    return (value);
}\
'''

DECODER_READ_INT16 = '''
static int16_t decoder_read_int16(struct decoder_t *self_p)
{
    int16_t value;

    value = (int16_t)decoder_read_uint16(self_p);
    value -= 32768;

    return (value);
}\
'''

DECODER_READ_INT32 = '''
static int32_t decoder_read_int32(struct decoder_t *self_p)
{
    int32_t value;

    value = (int32_t)decoder_read_uint32(self_p);
    value -= 2147483648;

    return (value);
}\
'''

DECODER_READ_INT64 = '''
static int64_t decoder_read_int64(struct decoder_t *self_p)
{
    uint64_t value;

    value = decoder_read_uint64(self_p);
    value -= 9223372036854775808ull;

    return ((int64_t)value);
}\
'''

DECODER_READ_BOOL = '''
static bool decoder_read_bool(struct decoder_t *self_p)
{
    return (decoder_read_bit(self_p) != 0);
}\
'''

DECODER_READ_NON_NEGATIVE_BINARY_INTEGER = '''
static uint64_t decoder_read_non_negative_binary_integer(struct decoder_t *self_p,
                                                         size_t size)
{
    size_t i;
    uint64_t value;

    value = 0;

    for (i = 0; i < size; i++) {
        value <<= 1;
        value |= (uint64_t)decoder_read_bit(self_p);
    }

    return (value);
}\
'''

functions = [
    (
        'decoder_read_non_negative_binary_integer(',
        DECODER_READ_NON_NEGATIVE_BINARY_INTEGER
    ),
    ('decoder_read_bool(', DECODER_READ_BOOL),
    ('decoder_read_int64(', DECODER_READ_INT64),
    ('decoder_read_int32(', DECODER_READ_INT32),
    ('decoder_read_int16(', DECODER_READ_INT16),
    ('decoder_read_int8(', DECODER_READ_INT8),
    ('decoder_read_uint64(', DECODER_READ_UINT64),
    ('decoder_read_uint32(', DECODER_READ_UINT32),
    ('decoder_read_uint16(', DECODER_READ_UINT16),
    ('decoder_read_uint8(', DECODER_READ_UINT8),
    ('decoder_read_bytes(', DECODER_READ_BYTES),
    ('decoder_read_bit(', DECODER_READ_BIT),
    ('decoder_free(', DECODER_FREE),
    ('decoder_abort(', DECODER_ABORT),
    ('decoder_get_result(', DECODER_GET_RESULT),
    ('decoder_init(', DECODER_INIT),
    (
        'encoder_append_non_negative_binary_integer(',
        ENCODER_APPEND_NON_NEGATIVE_BINARY_INTEGER
    ),
    ('encoder_append_bool(', ENCODER_APPEND_BOOL),
    ('encoder_append_int64(', ENCODER_APPEND_INT64),
    ('encoder_append_int32(', ENCODER_APPEND_INT32),
    ('encoder_append_int16(', ENCODER_APPEND_INT16),
    ('encoder_append_int8(', ENCODER_APPEND_INT8),
    ('encoder_append_uint64(', ENCODER_APPEND_UINT64),
    ('encoder_append_uint32(', ENCODER_APPEND_UINT32),
    ('encoder_append_uint16(', ENCODER_APPEND_UINT16),
    ('encoder_append_uint8(', ENCODER_APPEND_UINT8),
    ('encoder_append_bytes(', ENCODER_APPEND_BYTES),
    ('encoder_append_bit(', ENCODER_APPEND_BIT),
    ('encoder_alloc(', ENCODER_ALLOC),
    ('encoder_abort(', ENCODER_ABORT),
    ('encoder_get_result(', ENCODER_GET_RESULT),
    ('encoder_init(', ENCODER_INIT)
]



EXPECTED = {'Foo': {'extensibility-implied': False,
         'imports': {},
         'object-classes': {},
         'object-sets': {},
         'tags': 'AUTOMATIC',
         'types': {'A': {'restricted-to': [('min', 'max')],
                         'type': 'Constants'},
                   'B': {'restricted-to': ['unknown'], 'type': 'Constants'},
                   'C': {'restricted-to': [('zero', 'max')],
                         'type': 'Constants'},
                   'Constants': {'named-numbers': {'max': 1,
                                                   'min': -1,
                                                   'unknown': 2},
                                 'restricted-to': [(-1, 2)],
                                 'type': 'INTEGER'}},
         'values': {'min': {'type': 'INTEGER', 'value': 3},
                    'zero': {'type': 'INTEGER', 'value': 0}}}}


EXPECTED = {'ITS-Container': {'extensibility-implied': False,
                   'imports': {},
                   'object-classes': {},
                   'object-sets': {},
                   'tags': 'AUTOMATIC',
                   'types': {'AccelerationConfidence': {'named-numbers': {'outOfRange': 101,
                                                                          'pointOneMeterPerSecSquared': 1,
                                                                          'unavailable': 102},
                                                        'restricted-to': [(0,
                                                                           102)],
                                                        'type': 'INTEGER'},
                             'AccelerationControl': {'named-bits': [('brakePedalEngaged',
                                                                     '0'),
                                                                    ('gasPedalEngaged',
                                                                     '1'),
                                                                    ('emergencyBrakeEngaged',
                                                                     '2'),
                                                                    ('collisionWarningEngaged',
                                                                     '3'),
                                                                    ('accEngaged',
                                                                     '4'),
                                                                    ('cruiseControlEngaged',
                                                                     '5'),
                                                                    ('speedLimiterEngaged',
                                                                     '6')],
                                                     'size': [7],
                                                     'type': 'BIT STRING'},
                             'AccidentSubCauseCode': {'named-numbers': {'accidentInvolvingBus': 4,
                                                                        'accidentInvolvingHazardousMaterials': 5,
                                                                        'accidentInvolvingLorry': 3,
                                                                        'accidentOnOppositeLane': 6,
                                                                        'assistanceRequested': 8,
                                                                        'heavyAccident': 2,
                                                                        'multiVehicleAccident': 1,
                                                                        'unavailable': 0,
                                                                        'unsecuredAccident': 7},
                                                      'restricted-to': [(0,
                                                                         255)],
                                                      'type': 'INTEGER'},
                             'ActionID': {'members': [{'name': 'originatingStationID',
                                                       'type': 'StationID'},
                                                      {'name': 'sequenceNumber',
                                                       'type': 'SequenceNumber'}],
                                          'type': 'SEQUENCE'},
                             'AdverseWeatherCondition-AdhesionSubCauseCode': {'named-numbers': {'blackIceOnRoad': 6,
                                                                                                'fuelOnRoad': 2,
                                                                                                'heavyFrostOnRoad': 1,
                                                                                                'iceOnRoad': 5,
                                                                                                'instantBlackIce': 9,
                                                                                                'looseChippings': 8,
                                                                                                'mudOnRoad': 3,
                                                                                                'oilOnRoad': 7,
                                                                                                'roadsSalted': 10,
                                                                                                'snowOnRoad': 4,
                                                                                                'unavailable': 0},
                                                                              'restricted-to': [(0,
                                                                                                 255)],
                                                                              'type': 'INTEGER'},
                             'AdverseWeatherCondition-ExtremeWeatherConditionSubCauseCode': {'named-numbers': {'blizzard': 6,
                                                                                                               'damagingHail': 2,
                                                                                                               'hurricane': 3,
                                                                                                               'strongWinds': 1,
                                                                                                               'thunderstorm': 4,
                                                                                                               'tornado': 5,
                                                                                                               'unavailable': 0},
                                                                                             'restricted-to': [(0,
                                                                                                                255)],
                                                                                             'type': 'INTEGER'},
                             'AdverseWeatherCondition-PrecipitationSubCauseCode': {'named-numbers': {'heavyRain': 1,
                                                                                                     'heavySnowfall': 2,
                                                                                                     'softHail': 3,
                                                                                                     'unavailable': 0},
                                                                                   'restricted-to': [(0,
                                                                                                      255)],
                                                                                   'type': 'INTEGER'},
                             'AdverseWeatherCondition-VisibilitySubCauseCode': {'named-numbers': {'fog': 1,
                                                                                                  'heavyHail': 5,
                                                                                                  'heavyRain': 4,
                                                                                                  'heavySnowfall': 3,
                                                                                                  'lowSunGlare': 6,
                                                                                                  'sandstorms': 7,
                                                                                                  'smoke': 2,
                                                                                                  'swarmsOfInsects': 8,
                                                                                                  'unavailable': 0},
                                                                                'restricted-to': [(0,
                                                                                                   255)],
                                                                                'type': 'INTEGER'},
                             'Altitude': {'members': [{'name': 'altitudeValue',
                                                       'type': 'AltitudeValue'},
                                                      {'name': 'altitudeConfidence',
                                                       'type': 'AltitudeConfidence'}],
                                          'type': 'SEQUENCE'},
                             'AltitudeConfidence': {'type': 'ENUMERATED',
                                                    'values': [('alt-000-01',
                                                                0),
                                                               ('alt-000-02',
                                                                1),
                                                               ('alt-000-05',
                                                                2),
                                                               ('alt-000-10',
                                                                3),
                                                               ('alt-000-20',
                                                                4),
                                                               ('alt-000-50',
                                                                5),
                                                               ('alt-001-00',
                                                                6),
                                                               ('alt-002-00',
                                                                7),
                                                               ('alt-005-00',
                                                                8),
                                                               ('alt-010-00',
                                                                9),
                                                               ('alt-020-00',
                                                                10),
                                                               ('alt-050-00',
                                                                11),
                                                               ('alt-100-00',
                                                                12),
                                                               ('alt-200-00',
                                                                13),
                                                               ('outOfRange',
                                                                14),
                                                               ('unavailable',
                                                                15)]},
                             'AltitudeValue': {'named-numbers': {'oneCentimeter': 1,
                                                                 'referenceEllipsoidSurface': 0,
                                                                 'unavailable': 800001},
                                               'restricted-to': [(-100000,
                                                                  800001)],
                                               'type': 'INTEGER'},
                             'CauseCode': {'members': [{'name': 'causeCode',
                                                        'type': 'CauseCodeType'},
                                                       {'name': 'subCauseCode',
                                                        'type': 'SubCauseCodeType'}],
                                           'type': 'SEQUENCE'},
                             'CauseCodeType': {'named-numbers': {'accident': 2,
                                                                 'adverseWeatherCondition-Adhesion': 6,
                                                                 'adverseWeatherCondition-ExtremeWeatherCondition': 17,
                                                                 'adverseWeatherCondition-Precipitation': 19,
                                                                 'adverseWeatherCondition-Visibility': 18,
                                                                 'collisionRisk': 97,
                                                                 'dangerousEndOfQueue': 27,
                                                                 'dangerousSituation': 99,
                                                                 'emergencyVehicleApproaching': 95,
                                                                 'hazardousLocation-AnimalOnTheRoad': 11,
                                                                 'hazardousLocation-DangerousCurve': 96,
                                                                 'hazardousLocation-ObstacleOnTheRoad': 10,
                                                                 'hazardousLocation-SurfaceCondition': 9,
                                                                 'humanPresenceOnTheRoad': 12,
                                                                 'humanProblem': 93,
                                                                 'postCrash': 92,
                                                                 'rescueAndRecoveryWorkInProgress': 15,
                                                                 'reserved': 0,
                                                                 'roadworks': 3,
                                                                 'signalViolation': 98,
                                                                 'slowVehicle': 26,
                                                                 'stationaryVehicle': 94,
                                                                 'trafficCondition': 1,
                                                                 'vehicleBreakdown': 91,
                                                                 'wrongWayDriving': 14},
                                               'restricted-to': [(0, 255)],
                                               'type': 'INTEGER'},
                             'CenDsrcTollingZone': {'members': [{'name': 'protectedZoneLatitude',
                                                                 'type': 'Latitude'},
                                                                {'name': 'protectedZoneLongitude',
                                                                 'type': 'Longitude'},
                                                                {'name': 'cenDsrcTollingZoneID',
                                                                 'optional': True,
                                                                 'type': 'CenDsrcTollingZoneID'}],
                                                    'type': 'SEQUENCE'},
                             'CenDsrcTollingZoneID': {'type': 'ProtectedZoneID'},
                             'ClosedLanes': {'members': [{'name': 'hardShoulderStatus',
                                                          'optional': True,
                                                          'type': 'HardShoulderStatus'},
                                                         {'name': 'drivingLaneStatus',
                                                          'type': 'DrivingLaneStatus'},
                                                         None],
                                             'type': 'SEQUENCE'},
                             'CollisionRiskSubCauseCode': {'named-numbers': {'crossingCollisionRisk': 2,
                                                                             'lateralCollisionRisk': 3,
                                                                             'longitudinalCollisionRisk': 1,
                                                                             'unavailable': 0,
                                                                             'vulnerableRoadUser': 4},
                                                           'restricted-to': [(0,
                                                                              255)],
                                                           'type': 'INTEGER'},
                             'Curvature': {'members': [{'name': 'curvatureValue',
                                                        'type': 'CurvatureValue'},
                                                       {'name': 'curvatureConfidence',
                                                        'type': 'CurvatureConfidence'}],
                                           'type': 'SEQUENCE'},
                             'CurvatureCalculationMode': {'type': 'ENUMERATED',
                                                          'values': [('yawRateUsed',
                                                                      0),
                                                                     ('yawRateNotUsed',
                                                                      1),
                                                                     ('unavailable',
                                                                      2),
                                                                     None]},
                             'CurvatureConfidence': {'type': 'ENUMERATED',
                                                     'values': [('onePerMeter-0-00002',
                                                                 0),
                                                                ('onePerMeter-0-0001',
                                                                 1),
                                                                ('onePerMeter-0-0005',
                                                                 2),
                                                                ('onePerMeter-0-002',
                                                                 3),
                                                                ('onePerMeter-0-01',
                                                                 4),
                                                                ('onePerMeter-0-1',
                                                                 5),
                                                                ('outOfRange',
                                                                 6),
                                                                ('unavailable',
                                                                 7)]},
                             'CurvatureValue': {'named-numbers': {'reciprocalOf1MeterRadiusToLeft': 30000,
                                                                  'reciprocalOf1MeterRadiusToRight': -30000,
                                                                  'straight': 0,
                                                                  'unavailable': 30001},
                                                'restricted-to': [(-30000,
                                                                   30001)],
                                                'type': 'INTEGER'},
                             'DangerousEndOfQueueSubCauseCode': {'named-numbers': {'queueAroundBend': 3,
                                                                                   'queueInTunnel': 4,
                                                                                   'queueOverHill': 2,
                                                                                   'suddenEndOfQueue': 1,
                                                                                   'unavailable': 0},
                                                                 'restricted-to': [(0,
                                                                                    255)],
                                                                 'type': 'INTEGER'},
                             'DangerousGoodsBasic': {'type': 'ENUMERATED',
                                                     'values': [('explosives1',
                                                                 0),
                                                                ('explosives2',
                                                                 1),
                                                                ('explosives3',
                                                                 2),
                                                                ('explosives4',
                                                                 3),
                                                                ('explosives5',
                                                                 4),
                                                                ('explosives6',
                                                                 5),
                                                                ('flammableGases',
                                                                 6),
                                                                ('nonFlammableGases',
                                                                 7),
                                                                ('toxicGases',
                                                                 8),
                                                                ('flammableLiquids',
                                                                 9),
                                                                ('flammableSolids',
                                                                 10),
                                                                ('substancesLiableToSpontaneousCombustion',
                                                                 11),
                                                                ('substancesEmittingFlammableGasesUponContactWithWater',
                                                                 12),
                                                                ('oxidizingSubstances',
                                                                 13),
                                                                ('organicPeroxides',
                                                                 14),
                                                                ('toxicSubstances',
                                                                 15),
                                                                ('infectiousSubstances',
                                                                 16),
                                                                ('radioactiveMaterial',
                                                                 17),
                                                                ('corrosiveSubstances',
                                                                 18),
                                                                ('miscellaneousDangerousSubstances',
                                                                 19)]},
                             'DangerousGoodsExtended': {'members': [{'name': 'dangerousGoodsType',
                                                                     'type': 'DangerousGoodsBasic'},
                                                                    {'name': 'unNumber',
                                                                     'restricted-to': [(0,
                                                                                        9999)],
                                                                     'type': 'INTEGER'},
                                                                    {'name': 'elevatedTemperature',
                                                                     'type': 'BOOLEAN'},
                                                                    {'name': 'tunnelsRestricted',
                                                                     'type': 'BOOLEAN'},
                                                                    {'name': 'limitedQuantity',
                                                                     'type': 'BOOLEAN'},
                                                                    {'name': 'emergencyActionCode',
                                                                     'optional': True,
                                                                     'size': [(1,
                                                                               24)],
                                                                     'type': 'IA5String'},
                                                                    {'name': 'phoneNumber',
                                                                     'optional': True,
                                                                     'size': [(1,
                                                                               24)],
                                                                     'type': 'IA5String'},
                                                                    {'name': 'companyName',
                                                                     'optional': True,
                                                                     'size': [(1,
                                                                               24)],
                                                                     'type': 'UTF8String'}],
                                                        'type': 'SEQUENCE'},
                             'DangerousSituationSubCauseCode': {'named-numbers': {'absEngaged': 4,
                                                                                  'aebEngaged': 5,
                                                                                  'brakeWarningEngaged': 6,
                                                                                  'collisionRiskWarningEngaged': 7,
                                                                                  'emergencyElectronicBrakeEngaged': 1,
                                                                                  'espEngaged': 3,
                                                                                  'preCrashSystemEngaged': 2,
                                                                                  'unavailable': 0},
                                                                'restricted-to': [(0,
                                                                                   255)],
                                                                'type': 'INTEGER'},
                             'DeltaAltitude': {'named-numbers': {'oneCentimeterDown': -1,
                                                                 'oneCentimeterUp': 1,
                                                                 'unavailable': 12800},
                                               'restricted-to': [(-12700,
                                                                  12800)],
                                               'type': 'INTEGER'},
                             'DeltaLatitude': {'named-numbers': {'oneMicrodegreeNorth': 10,
                                                                 'oneMicrodegreeSouth': -10,
                                                                 'unavailable': 131072},
                                               'restricted-to': [(-131071,
                                                                  131072)],
                                               'type': 'INTEGER'},
                             'DeltaLongitude': {'named-numbers': {'oneMicrodegreeEast': 10,
                                                                  'oneMicrodegreeWest': -10,
                                                                  'unavailable': 131072},
                                                'restricted-to': [(-131071,
                                                                   131072)],
                                                'type': 'INTEGER'},
                             'DeltaReferencePosition': {'members': [{'name': 'deltaLatitude',
                                                                     'type': 'DeltaLatitude'},
                                                                    {'name': 'deltaLongitude',
                                                                     'type': 'DeltaLongitude'},
                                                                    {'name': 'deltaAltitude',
                                                                     'type': 'DeltaAltitude'}],
                                                        'type': 'SEQUENCE'},
                             'DriveDirection': {'type': 'ENUMERATED',
                                                'values': [('forward', 0),
                                                           ('backward', 1),
                                                           ('unavailable', 2)]},
                             'DrivingLaneStatus': {'named-bits': [('outermostLaneClosed',
                                                                   '1'),
                                                                  ('secondLaneFromOutsideClosed',
                                                                   '2')],
                                                   'size': [(1, 14)],
                                                   'type': 'BIT STRING'},
                             'EmbarkationStatus': {'type': 'BOOLEAN'},
                             'EmergencyPriority': {'named-bits': [('requestForRightOfWay',
                                                                   '0'),
                                                                  ('requestForFreeCrossingAtATrafficLight',
                                                                   '1')],
                                                   'size': [2],
                                                   'type': 'BIT STRING'},
                             'EmergencyVehicleApproachingSubCauseCode': {'named-numbers': {'emergencyVehicleApproaching': 1,
                                                                                           'prioritizedVehicleApproaching': 2,
                                                                                           'unavailable': 0},
                                                                         'restricted-to': [(0,
                                                                                            255)],
                                                                         'type': 'INTEGER'},
                             'EnergyStorageType': {'named-bits': [('hydrogenStorage',
                                                                   '0'),
                                                                  ('electricEnergyStorage',
                                                                   '1'),
                                                                  ('liquidPropaneGas',
                                                                   '2'),
                                                                  ('compressedNaturalGas',
                                                                   '3'),
                                                                  ('diesel',
                                                                   '4'),
                                                                  ('gasoline',
                                                                   '5'),
                                                                  ('ammonia',
                                                                   '6')],
                                                   'size': [7],
                                                   'type': 'BIT STRING'},
                             'EventHistory': {'element': {'type': 'EventPoint'},
                                              'size': [(1, 23)],
                                              'type': 'SEQUENCE OF'},
                             'EventPoint': {'members': [{'name': 'eventPosition',
                                                         'type': 'DeltaReferencePosition'},
                                                        {'name': 'eventDeltaTime',
                                                         'optional': True,
                                                         'type': 'PathDeltaTime'},
                                                        {'name': 'informationQuality',
                                                         'type': 'InformationQuality'}],
                                            'type': 'SEQUENCE'},
                             'ExteriorLights': {'named-bits': [('lowBeamHeadlightsOn',
                                                                '0'),
                                                               ('highBeamHeadlightsOn',
                                                                '1'),
                                                               ('leftTurnSignalOn',
                                                                '2'),
                                                               ('rightTurnSignalOn',
                                                                '3'),
                                                               ('daytimeRunningLightsOn',
                                                                '4'),
                                                               ('reverseLightOn',
                                                                '5'),
                                                               ('fogLightOn',
                                                                '6'),
                                                               ('parkingLightsOn',
                                                                '7')],
                                                'size': [8],
                                                'type': 'BIT STRING'},
                             'HardShoulderStatus': {'type': 'ENUMERATED',
                                                    'values': [('availableForStopping',
                                                                0),
                                                               ('closed', 1),
                                                               ('availableForDriving',
                                                                2)]},
                             'HazardousLocation-AnimalOnTheRoadSubCauseCode': {'named-numbers': {'herdOfAnimals': 2,
                                                                                                 'largeAnimals': 4,
                                                                                                 'smallAnimals': 3,
                                                                                                 'unavailable': 0,
                                                                                                 'wildAnimals': 1},
                                                                               'restricted-to': [(0,
                                                                                                  255)],
                                                                               'type': 'INTEGER'},
                             'HazardousLocation-DangerousCurveSubCauseCode': {'named-numbers': {'dangerousLeftTurnCurve': 1,
                                                                                                'dangerousRightTurnCurve': 2,
                                                                                                'multipleCurvesStartingWithLeftTurn': 4,
                                                                                                'multipleCurvesStartingWithRightTurn': 5,
                                                                                                'multipleCurvesStartingWithUnknownTurningDirection': 3,
                                                                                                'unavailable': 0},
                                                                              'restricted-to': [(0,
                                                                                                 255)],
                                                                              'type': 'INTEGER'},
                             'HazardousLocation-ObstacleOnTheRoadSubCauseCode': {'named-numbers': {'bigObjects': 4,
                                                                                                   'fallenTrees': 5,
                                                                                                   'hubCaps': 6,
                                                                                                   'partsOfTyres': 3,
                                                                                                   'partsOfVehicles': 2,
                                                                                                   'shedLoad': 1,
                                                                                                   'unavailable': 0,
                                                                                                   'waitingVehicles': 7},
                                                                                 'restricted-to': [(0,
                                                                                                    255)],
                                                                                 'type': 'INTEGER'},
                             'HazardousLocation-SurfaceConditionSubCauseCode': {'named-numbers': {'burstPipe': 7,
                                                                                                  'earthquakeDamage': 2,
                                                                                                  'fallingIce': 9,
                                                                                                  'rockfalls': 1,
                                                                                                  'sewerCollapse': 3,
                                                                                                  'snowDrifts': 5,
                                                                                                  'stormDamage': 6,
                                                                                                  'subsidence': 4,
                                                                                                  'unavailable': 0,
                                                                                                  'volcanoEruption': 8},
                                                                                'restricted-to': [(0,
                                                                                                   255)],
                                                                                'type': 'INTEGER'},
                             'Heading': {'members': [{'name': 'headingValue',
                                                      'type': 'HeadingValue'},
                                                     {'name': 'headingConfidence',
                                                      'type': 'HeadingConfidence'}],
                                         'type': 'SEQUENCE'},
                             'HeadingConfidence': {'named-numbers': {'equalOrWithinOneDegree': 10,
                                                                     'equalOrWithinZeroPointOneDegree': 1,
                                                                     'outOfRange': 126,
                                                                     'unavailable': 127},
                                                   'restricted-to': [(1, 127)],
                                                   'type': 'INTEGER'},
                             'HeadingValue': {'named-numbers': {'unavailable': 3601,
                                                                'wgs84East': 900,
                                                                'wgs84North': 0,
                                                                'wgs84South': 1800,
                                                                'wgs84West': 2700},
                                              'restricted-to': [(0, 3601)],
                                              'type': 'INTEGER'},
                             'HeightLonCarr': {'named-numbers': {'oneCentimeter': 1,
                                                                 'unavailable': 100},
                                               'restricted-to': [(1, 100)],
                                               'type': 'INTEGER'},
                             'HumanPresenceOnTheRoadSubCauseCode': {'named-numbers': {'childrenOnRoadway': 1,
                                                                                      'cyclistOnRoadway': 2,
                                                                                      'motorcyclistOnRoadway': 3,
                                                                                      'unavailable': 0},
                                                                    'restricted-to': [(0,
                                                                                       255)],
                                                                    'type': 'INTEGER'},
                             'HumanProblemSubCauseCode': {'named-numbers': {'glycemiaProblem': 1,
                                                                            'heartProblem': 2,
                                                                            'unavailable': 0},
                                                          'restricted-to': [(0,
                                                                             255)],
                                                          'type': 'INTEGER'},
                             'InformationQuality': {'named-numbers': {'highest': 7,
                                                                      'lowest': 1,
                                                                      'unavailable': 0},
                                                    'restricted-to': [(0, 7)],
                                                    'type': 'INTEGER'},
                             'ItineraryPath': {'element': {'type': 'ReferencePosition'},
                                               'size': [(1, 40)],
                                               'type': 'SEQUENCE OF'},
                             'ItsPduHeader': {'members': [{'name': 'protocolVersion',
                                                           'named-numbers': {'currentVersion': 1},
                                                           'restricted-to': [(0,
                                                                              255)],
                                                           'type': 'INTEGER'},
                                                          {'name': 'messageID',
                                                           'named-numbers': {'cam': 2,
                                                                             'denm': 1,
                                                                             'ev-rsr': 7,
                                                                             'ivi': 6,
                                                                             'map': 5,
                                                                             'poi': 3,
                                                                             'spat': 4},
                                                           'restricted-to': [(0,
                                                                              255)],
                                                           'type': 'INTEGER'},
                                                          {'name': 'stationID',
                                                           'type': 'StationID'}],
                                              'type': 'SEQUENCE'},
                             'LanePosition': {'named-numbers': {'hardShoulder': 0,
                                                                'offTheRoad': -1,
                                                                'outermostDrivingLane': 1,
                                                                'secondLaneFromOutside': 2},
                                              'restricted-to': [(-1, 14)],
                                              'type': 'INTEGER'},
                             'LateralAcceleration': {'members': [{'name': 'lateralAccelerationValue',
                                                                  'type': 'LateralAccelerationValue'},
                                                                 {'name': 'lateralAccelerationConfidence',
                                                                  'type': 'AccelerationConfidence'}],
                                                     'type': 'SEQUENCE'},
                             'LateralAccelerationValue': {'named-numbers': {'pointOneMeterPerSecSquaredToLeft': 1,
                                                                            'pointOneMeterPerSecSquaredToRight': -1,
                                                                            'unavailable': 161},
                                                          'restricted-to': [(-160,
                                                                             161)],
                                                          'type': 'INTEGER'},
                             'Latitude': {'named-numbers': {'oneMicrodegreeNorth': 10,
                                                            'oneMicrodegreeSouth': -10,
                                                            'unavailable': 900000001},
                                          'restricted-to': [(-900000000,
                                                             900000001)],
                                          'type': 'INTEGER'},
                             'LightBarSirenInUse': {'named-bits': [('lightBarActivated',
                                                                    '0'),
                                                                   ('sirenActivated',
                                                                    '1')],
                                                    'size': [2],
                                                    'type': 'BIT STRING'},
                             'Longitude': {'named-numbers': {'oneMicrodegreeEast': 10,
                                                             'oneMicrodegreeWest': -10,
                                                             'unavailable': 1800000001},
                                           'restricted-to': [(-1800000000,
                                                              1800000001)],
                                           'type': 'INTEGER'},
                             'LongitudinalAcceleration': {'members': [{'name': 'longitudinalAccelerationValue',
                                                                       'type': 'LongitudinalAccelerationValue'},
                                                                      {'name': 'longitudinalAccelerationConfidence',
                                                                       'type': 'AccelerationConfidence'}],
                                                          'type': 'SEQUENCE'},
                             'LongitudinalAccelerationValue': {'named-numbers': {'pointOneMeterPerSecSquaredBackward': -1,
                                                                                 'pointOneMeterPerSecSquaredForward': 1,
                                                                                 'unavailable': 161},
                                                               'restricted-to': [(-160,
                                                                                  161)],
                                                               'type': 'INTEGER'},
                             'NumberOfOccupants': {'named-numbers': {'oneOccupant': 1,
                                                                     'unavailable': 127},
                                                   'restricted-to': [(0, 127)],
                                                   'type': 'INTEGER'},
                             'PathDeltaTime': {'named-numbers': {'tenMilliSecondsInPast': 1},
                                               'restricted-to': [(1, 65535),
                                                                 None],
                                               'type': 'INTEGER'},
                             'PathHistory': {'element': {'type': 'PathPoint'},
                                             'size': [(0, 40)],
                                             'type': 'SEQUENCE OF'},
                             'PathPoint': {'members': [{'name': 'pathPosition',
                                                        'type': 'DeltaReferencePosition'},
                                                       {'name': 'pathDeltaTime',
                                                        'optional': True,
                                                        'type': 'PathDeltaTime'}],
                                           'type': 'SEQUENCE'},
                             'PerformanceClass': {'named-numbers': {'performanceClassA': 1,
                                                                    'performanceClassB': 2,
                                                                    'unavailable': 0},
                                                  'restricted-to': [(0, 7)],
                                                  'type': 'INTEGER'},
                             'PosCentMass': {'named-numbers': {'tenCentimeters': 1,
                                                               'unavailable': 63},
                                             'restricted-to': [(1, 63)],
                                             'type': 'INTEGER'},
                             'PosConfidenceEllipse': {'members': [{'name': 'semiMajorConfidence',
                                                                   'type': 'SemiAxisLength'},
                                                                  {'name': 'semiMinorConfidence',
                                                                   'type': 'SemiAxisLength'},
                                                                  {'name': 'semiMajorOrientation',
                                                                   'type': 'HeadingValue'}],
                                                      'type': 'SEQUENCE'},
                             'PosFrontAx': {'named-numbers': {'tenCentimeters': 1,
                                                              'unavailable': 20},
                                            'restricted-to': [(1, 20)],
                                            'type': 'INTEGER'},
                             'PosLonCarr': {'named-numbers': {'oneCentimeter': 1,
                                                              'unavailable': 127},
                                            'restricted-to': [(1, 127)],
                                            'type': 'INTEGER'},
                             'PosPillar': {'named-numbers': {'tenCentimeters': 1,
                                                             'unavailable': 30},
                                           'restricted-to': [(1, 30)],
                                           'type': 'INTEGER'},
                             'PositionOfOccupants': {'named-bits': [('row1LeftOccupied',
                                                                     '0'),
                                                                    ('row1RightOccupied',
                                                                     '1'),
                                                                    ('row1MidOccupied',
                                                                     '2'),
                                                                    ('row1NotDetectable',
                                                                     '3'),
                                                                    ('row1NotPresent',
                                                                     '4'),
                                                                    ('row2LeftOccupied',
                                                                     '5'),
                                                                    ('row2RightOccupied',
                                                                     '6'),
                                                                    ('row2MidOccupied',
                                                                     '7'),
                                                                    ('row2NotDetectable',
                                                                     '8'),
                                                                    ('row2NotPresent',
                                                                     '9'),
                                                                    ('row3LeftOccupied',
                                                                     '10'),
                                                                    ('row3RightOccupied',
                                                                     '11'),
                                                                    ('row3MidOccupied',
                                                                     '12'),
                                                                    ('row3NotDetectable',
                                                                     '13'),
                                                                    ('row3NotPresent',
                                                                     '14'),
                                                                    ('row4LeftOccupied',
                                                                     '15'),
                                                                    ('row4RightOccupied',
                                                                     '16'),
                                                                    ('row4MidOccupied',
                                                                     '17'),
                                                                    ('row4NotDetectable',
                                                                     '18'),
                                                                    ('row4NotPresent',
                                                                     '19')],
                                                     'size': [20],
                                                     'type': 'BIT STRING'},
                             'PositionOfPillars': {'element': {'type': 'PosPillar'},
                                                   'size': [(1, 3), None],
                                                   'type': 'SEQUENCE OF'},
                             'PositioningSolutionType': {'type': 'ENUMERATED',
                                                         'values': [('noPositioningSolution',
                                                                     0),
                                                                    ('sGNSS',
                                                                     1),
                                                                    ('dGNSS',
                                                                     2),
                                                                    ('sGNSSplusDR',
                                                                     3),
                                                                    ('dGNSSplusDR',
                                                                     4),
                                                                    ('dR', 5),
                                                                    None]},
                             'PostCrashSubCauseCode': {'named-numbers': {'accidentWithECallAutomaticallyTriggered': 3,
                                                                         'accidentWithECallManuallyTriggered': 2,
                                                                         'accidentWithECallTriggeredWithoutAccessToCellularNetwork': 4,
                                                                         'accidentWithoutECallTriggered': 1,
                                                                         'unavailable': 0},
                                                       'restricted-to': [(0,
                                                                          255)],
                                                       'type': 'INTEGER'},
                             'ProtectedCommunicationZone': {'members': [{'name': 'protectedZoneType',
                                                                         'type': 'ProtectedZoneType'},
                                                                        {'name': 'expiryTime',
                                                                         'optional': True,
                                                                         'type': 'TimestampIts'},
                                                                        {'name': 'protectedZoneLatitude',
                                                                         'type': 'Latitude'},
                                                                        {'name': 'protectedZoneLongitude',
                                                                         'type': 'Longitude'},
                                                                        {'name': 'protectedZoneRadius',
                                                                         'optional': True,
                                                                         'type': 'ProtectedZoneRadius'},
                                                                        {'name': 'protectedZoneID',
                                                                         'optional': True,
                                                                         'type': 'ProtectedZoneID'}],
                                                            'type': 'SEQUENCE'},
                             'ProtectedCommunicationZonesRSU': {'element': {'type': 'ProtectedCommunicationZone'},
                                                                'size': [(1,
                                                                          16)],
                                                                'type': 'SEQUENCE '
                                                                        'OF'},
                             'ProtectedZoneID': {'restricted-to': [(0,
                                                                    134217727)],
                                                 'type': 'INTEGER'},
                             'ProtectedZoneRadius': {'named-numbers': {'oneMeter': 1},
                                                     'restricted-to': [(1, 255),
                                                                       None],
                                                     'type': 'INTEGER'},
                             'ProtectedZoneType': {'type': 'ENUMERATED',
                                                   'values': [('cenDsrcTolling',
                                                               0),
                                                              None]},
                             'PtActivation': {'members': [{'name': 'ptActivationType',
                                                           'type': 'PtActivationType'},
                                                          {'name': 'ptActivationData',
                                                           'type': 'PtActivationData'}],
                                              'type': 'SEQUENCE'},
                             'PtActivationData': {'size': [(1, 20)],
                                                  'type': 'OCTET STRING'},
                             'PtActivationType': {'named-numbers': {'r09-16CodingType': 1,
                                                                    'undefinedCodingType': 0,
                                                                    'vdv-50149CodingType': 2},
                                                  'restricted-to': [(0, 255)],
                                                  'type': 'INTEGER'},
                             'ReferencePosition': {'members': [{'name': 'latitude',
                                                                'type': 'Latitude'},
                                                               {'name': 'longitude',
                                                                'type': 'Longitude'},
                                                               {'name': 'positionConfidenceEllipse',
                                                                'type': 'PosConfidenceEllipse'},
                                                               {'name': 'altitude',
                                                                'type': 'Altitude'}],
                                                   'type': 'SEQUENCE'},
                             'RelevanceDistance': {'type': 'ENUMERATED',
                                                   'values': [('lessThan50m',
                                                               0),
                                                              ('lessThan100m',
                                                               1),
                                                              ('lessThan200m',
                                                               2),
                                                              ('lessThan500m',
                                                               3),
                                                              ('lessThan1000m',
                                                               4),
                                                              ('lessThan5km',
                                                               5),
                                                              ('lessThan10km',
                                                               6),
                                                              ('over10km', 7)]},
                             'RelevanceTrafficDirection': {'type': 'ENUMERATED',
                                                           'values': [('allTrafficDirections',
                                                                       0),
                                                                      ('upstreamTraffic',
                                                                       1),
                                                                      ('downstreamTraffic',
                                                                       2),
                                                                      ('oppositeTraffic',
                                                                       3)]},
                             'RequestResponseIndication': {'type': 'ENUMERATED',
                                                           'values': [('request',
                                                                       0),
                                                                      ('response',
                                                                       1)]},
                             'RescueAndRecoveryWorkInProgressSubCauseCode': {'named-numbers': {'childAbductionInProgress': 5,
                                                                                               'emergencyVehicles': 1,
                                                                                               'medicalEmergencyOngoing': 4,
                                                                                               'policeActivityOngoing': 3,
                                                                                               'rescueHelicopterLanding': 2,
                                                                                               'unavailable': 0},
                                                                             'restricted-to': [(0,
                                                                                                255)],
                                                                             'type': 'INTEGER'},
                             'RestrictedTypes': {'element': {'type': 'StationType'},
                                                 'size': [(1, 3), None],
                                                 'type': 'SEQUENCE OF'},
                             'RoadType': {'type': 'ENUMERATED',
                                          'values': [('urban-NoStructuralSeparationToOppositeLanes',
                                                      0),
                                                     ('urban-WithStructuralSeparationToOppositeLanes',
                                                      1),
                                                     ('nonUrban-NoStructuralSeparationToOppositeLanes',
                                                      2),
                                                     ('nonUrban-WithStructuralSeparationToOppositeLanes',
                                                      3)]},
                             'RoadworksSubCauseCode': {'named-numbers': {'majorRoadworks': 1,
                                                                         'roadMarkingWork': 2,
                                                                         'shortTermStationaryRoadworks': 4,
                                                                         'slowMovingRoadMaintenance': 3,
                                                                         'streetCleaning': 5,
                                                                         'unavailable': 0,
                                                                         'winterService': 6},
                                                       'restricted-to': [(0,
                                                                          255)],
                                                       'type': 'INTEGER'},
                             'SemiAxisLength': {'named-numbers': {'oneCentimeter': 1,
                                                                  'outOfRange': 4094,
                                                                  'unavailable': 4095},
                                                'restricted-to': [(0, 4095)],
                                                'type': 'INTEGER'},
                             'SequenceNumber': {'restricted-to': [(0, 65535)],
                                                'type': 'INTEGER'},
                             'SignalViolationSubCauseCode': {'named-numbers': {'stopSignViolation': 1,
                                                                               'trafficLightViolation': 2,
                                                                               'turningRegulationViolation': 3,
                                                                               'unavailable': 0},
                                                             'restricted-to': [(0,
                                                                                255)],
                                                             'type': 'INTEGER'},
                             'SlowVehicleSubCauseCode': {'named-numbers': {'abnormalLoad': 3,
                                                                           'abnormalWideLoad': 4,
                                                                           'convoy': 5,
                                                                           'deicing': 7,
                                                                           'maintenanceVehicle': 1,
                                                                           'saltingVehicles': 8,
                                                                           'snowplough': 6,
                                                                           'unavailable': 0,
                                                                           'vehiclesSlowingToLookAtAccident': 2},
                                                         'restricted-to': [(0,
                                                                            255)],
                                                         'type': 'INTEGER'},
                             'SpecialTransportType': {'named-bits': [('heavyLoad',
                                                                      '0'),
                                                                     ('excessWidth',
                                                                      '1'),
                                                                     ('excessLength',
                                                                      '2'),
                                                                     ('excessHeight',
                                                                      '3')],
                                                      'size': [4],
                                                      'type': 'BIT STRING'},
                             'Speed': {'members': [{'name': 'speedValue',
                                                    'type': 'SpeedValue'},
                                                   {'name': 'speedConfidence',
                                                    'type': 'SpeedConfidence'}],
                                       'type': 'SEQUENCE'},
                             'SpeedConfidence': {'named-numbers': {'equalOrWithinOneCentimeterPerSec': 1,
                                                                   'equalOrWithinOneMeterPerSec': 100,
                                                                   'outOfRange': 126,
                                                                   'unavailable': 127},
                                                 'restricted-to': [(1, 127)],
                                                 'type': 'INTEGER'},
                             'SpeedLimit': {'named-numbers': {'oneKmPerHour': 1},
                                            'restricted-to': [(1, 255)],
                                            'type': 'INTEGER'},
                             'SpeedValue': {'named-numbers': {'oneCentimeterPerSec': 1,
                                                              'standstill': 0,
                                                              'unavailable': 16383},
                                            'restricted-to': [(0, 16383)],
                                            'type': 'INTEGER'},
                             'StationID': {'restricted-to': [(0, 4294967295)],
                                           'type': 'INTEGER'},
                             'StationType': {'named-numbers': {'bus': 6,
                                                               'cyclist': 2,
                                                               'heavyTruck': 8,
                                                               'lightTruck': 7,
                                                               'moped': 3,
                                                               'motorcycle': 4,
                                                               'passengerCar': 5,
                                                               'pedestrian': 1,
                                                               'roadSideUnit': 15,
                                                               'specialVehicles': 10,
                                                               'trailer': 9,
                                                               'tram': 11,
                                                               'unknown': 0},
                                             'restricted-to': [(0, 255)],
                                             'type': 'INTEGER'},
                             'StationarySince': {'type': 'ENUMERATED',
                                                 'values': [('lessThan1Minute',
                                                             0),
                                                            ('lessThan2Minutes',
                                                             1),
                                                            ('lessThan15Minutes',
                                                             2),
                                                            ('equalOrGreater15Minutes',
                                                             3)]},
                             'StationaryVehicleSubCauseCode': {'named-numbers': {'carryingDangerousGoods': 5,
                                                                                 'humanProblem': 1,
                                                                                 'postCrash': 3,
                                                                                 'publicTransportStop': 4,
                                                                                 'unavailable': 0,
                                                                                 'vehicleBreakdown': 2},
                                                               'restricted-to': [(0,
                                                                                  255)],
                                                               'type': 'INTEGER'},
                             'SteeringWheelAngle': {'members': [{'name': 'steeringWheelAngleValue',
                                                                 'type': 'SteeringWheelAngleValue'},
                                                                {'name': 'steeringWheelAngleConfidence',
                                                                 'type': 'SteeringWheelAngleConfidence'}],
                                                    'type': 'SEQUENCE'},
                             'SteeringWheelAngleConfidence': {'named-numbers': {'equalOrWithinOnePointFiveDegree': 1,
                                                                                'outOfRange': 126,
                                                                                'unavailable': 127},
                                                              'restricted-to': [(1,
                                                                                 127)],
                                                              'type': 'INTEGER'},
                             'SteeringWheelAngleValue': {'named-numbers': {'onePointFiveDegreesToLeft': 1,
                                                                           'onePointFiveDegreesToRight': -1,
                                                                           'straight': 0,
                                                                           'unavailable': 512},
                                                         'restricted-to': [(-511,
                                                                            512)],
                                                         'type': 'INTEGER'},
                             'SubCauseCodeType': {'restricted-to': [(0, 255)],
                                                  'type': 'INTEGER'},
                             'Temperature': {'named-numbers': {'equalOrGreaterThan67Deg': 67,
                                                               'equalOrSmallerThanMinus60Deg': -60,
                                                               'oneDegreeCelsius': 1},
                                             'restricted-to': [(-60, 67)],
                                             'type': 'INTEGER'},
                             'TimestampIts': {'named-numbers': {'oneMillisecAfterUTCStartOf2004': 1,
                                                                'utcStartOf2004': 0},
                                              'restricted-to': [(0,
                                                                 4398046511103)],
                                              'type': 'INTEGER'},
                             'Traces': {'element': {'type': 'PathHistory'},
                                        'size': [(1, 7)],
                                        'type': 'SEQUENCE OF'},
                             'TrafficConditionSubCauseCode': {'named-numbers': {'increasedVolumeOfTraffic': 1,
                                                                                'trafficJamDecreasing': 7,
                                                                                'trafficJamIncreasing': 3,
                                                                                'trafficJamSlightlyDecreasing': 6,
                                                                                'trafficJamSlowlyIncreasing': 2,
                                                                                'trafficJamStronglyDecreasing': 8,
                                                                                'trafficJamStronglyIncreasing': 4,
                                                                                'trafficStationary': 5,
                                                                                'unavailable': 0},
                                                              'restricted-to': [(0,
                                                                                 255)],
                                                              'type': 'INTEGER'},
                             'TrafficRule': {'type': 'ENUMERATED',
                                             'values': [('noPassing', 0),
                                                        ('noPassingForTrucks',
                                                         1),
                                                        ('passToRight', 2),
                                                        ('passToLeft', 3),
                                                        None]},
                             'TransmissionInterval': {'named-numbers': {'oneMilliSecond': 1,
                                                                        'tenSeconds': 10000},
                                                      'restricted-to': [(1,
                                                                         10000)],
                                                      'type': 'INTEGER'},
                             'TurningRadius': {'named-numbers': {'point4Meters': 1,
                                                                 'unavailable': 255},
                                               'restricted-to': [(1, 255)],
                                               'type': 'INTEGER'},
                             'VDS': {'size': [6], 'type': 'IA5String'},
                             'ValidityDuration': {'named-numbers': {'oneSecondAfterDetection': 1,
                                                                    'timeOfDetection': 0},
                                                  'restricted-to': [(0, 86400)],
                                                  'type': 'INTEGER'},
                             'VehicleBreakdownSubCauseCode': {'named-numbers': {'brakingSystemProblem': 6,
                                                                                'engineCoolingProblem': 5,
                                                                                'engineProblem': 3,
                                                                                'lackOfBatteryPower': 2,
                                                                                'lackOfFuel': 1,
                                                                                'steeringProblem': 7,
                                                                                'transmissionProblem': 4,
                                                                                'tyrePuncture': 8,
                                                                                'unavailable': 0},
                                                              'restricted-to': [(0,
                                                                                 255)],
                                                              'type': 'INTEGER'},
                             'VehicleIdentification': {'members': [{'name': 'wMInumber',
                                                                    'optional': True,
                                                                    'type': 'WMInumber'},
                                                                   {'name': 'vDS',
                                                                    'optional': True,
                                                                    'type': 'VDS'},
                                                                   None],
                                                       'type': 'SEQUENCE'},
                             'VehicleLength': {'members': [{'name': 'vehicleLengthValue',
                                                            'type': 'VehicleLengthValue'},
                                                           {'name': 'vehicleLengthConfidenceIndication',
                                                            'type': 'VehicleLengthConfidenceIndication'}],
                                               'type': 'SEQUENCE'},
                             'VehicleLengthConfidenceIndication': {'type': 'ENUMERATED',
                                                                   'values': [('noTrailerPresent',
                                                                               0),
                                                                              ('trailerPresentWithKnownLength',
                                                                               1),
                                                                              ('trailerPresentWithUnknownLength',
                                                                               2),
                                                                              ('trailerPresenceIsUnknown',
                                                                               3),
                                                                              ('unavailable',
                                                                               4)]},
                             'VehicleLengthValue': {'named-numbers': {'outOfRange': 1022,
                                                                      'tenCentimeters': 1,
                                                                      'unavailable': 1023},
                                                    'restricted-to': [(1,
                                                                       1023)],
                                                    'type': 'INTEGER'},
                             'VehicleMass': {'named-numbers': {'hundredKg': 1,
                                                               'unavailable': 1024},
                                             'restricted-to': [(1, 1024)],
                                             'type': 'INTEGER'},
                             'VehicleRole': {'type': 'ENUMERATED',
                                             'values': [('default', 0),
                                                        ('publicTransport', 1),
                                                        ('specialTransport', 2),
                                                        ('dangerousGoods', 3),
                                                        ('roadWork', 4),
                                                        ('rescue', 5),
                                                        ('emergency', 6),
                                                        ('safetyCar', 7),
                                                        ('agriculture', 8),
                                                        ('commercial', 9),
                                                        ('military', 10),
                                                        ('roadOperator', 11),
                                                        ('taxi', 12),
                                                        ('reserved1', 13),
                                                        ('reserved2', 14),
                                                        ('reserved3', 15)]},
                             'VehicleWidth': {'named-numbers': {'outOfRange': 61,
                                                                'tenCentimeters': 1,
                                                                'unavailable': 62},
                                              'restricted-to': [(1, 62)],
                                              'type': 'INTEGER'},
                             'VerticalAcceleration': {'members': [{'name': 'verticalAccelerationValue',
                                                                   'type': 'VerticalAccelerationValue'},
                                                                  {'name': 'verticalAccelerationConfidence',
                                                                   'type': 'AccelerationConfidence'}],
                                                      'type': 'SEQUENCE'},
                             'VerticalAccelerationValue': {'named-numbers': {'pointOneMeterPerSecSquaredDown': -1,
                                                                             'pointOneMeterPerSecSquaredUp': 1,
                                                                             'unavailable': 161},
                                                           'restricted-to': [(-160,
                                                                              161)],
                                                           'type': 'INTEGER'},
                             'WMInumber': {'size': [(1, 3)],
                                           'type': 'IA5String'},
                             'WheelBaseVehicle': {'named-numbers': {'tenCentimeters': 1,
                                                                    'unavailable': 127},
                                                  'restricted-to': [(1, 127)],
                                                  'type': 'INTEGER'},
                             'WrongWayDrivingSubCauseCode': {'named-numbers': {'unavailable': 0,
                                                                               'wrongDirection': 2,
                                                                               'wrongLane': 1},
                                                             'restricted-to': [(0,
                                                                                255)],
                                                             'type': 'INTEGER'},
                             'YawRate': {'members': [{'name': 'yawRateValue',
                                                      'type': 'YawRateValue'},
                                                     {'name': 'yawRateConfidence',
                                                      'type': 'YawRateConfidence'}],
                                         'type': 'SEQUENCE'},
                             'YawRateConfidence': {'type': 'ENUMERATED',
                                                   'values': [('degSec-000-01',
                                                               0),
                                                              ('degSec-000-05',
                                                               1),
                                                              ('degSec-000-10',
                                                               2),
                                                              ('degSec-001-00',
                                                               3),
                                                              ('degSec-005-00',
                                                               4),
                                                              ('degSec-010-00',
                                                               5),
                                                              ('degSec-100-00',
                                                               6),
                                                              ('outOfRange', 7),
                                                              ('unavailable',
                                                               8)]},
                             'YawRateValue': {'named-numbers': {'degSec-000-01ToLeft': 1,
                                                                'degSec-000-01ToRight': -1,
                                                                'straight': 0,
                                                                'unavailable': 32767},
                                              'restricted-to': [(-32766,
                                                                 32767)],
                                              'type': 'INTEGER'}},
                   'values': {}}}


EXPECTED = {'Foo': {'extensibility-implied': False,
         'imports': {},
         'object-classes': {},
         'object-sets': {},
         'types': {'A': {'type': 'INTEGER'}, 'C': {'type': 'OCTET STRING'}},
         'values': {}}}


EXPECTED = {'Foo': {'extensibility-implied': False,
         'imports': {},
         'object-classes': {},
         'object-sets': {},
         'types': {'B': {'type': 'DATE'},
                   'C': {'type': 'TIME-OF-DAY'},
                   'D': {'type': 'DATE-TIME'}},
         'values': {}}}



import unittest

import asn1tools


CODECS_AND_MODULES = [
    ('oer', asn1tools.source.c.oer),
    ('uper', asn1tools.source.c.uper)
]


class Asn1ToolsCSourceTest(unittest.TestCase):

    def test_compile_error_unsupported_type(self):
        for codec, module in CODECS_AND_MODULES:
            foo = asn1tools.compile_string(
                'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
                '    A ::= OBJECT IDENTIFIER '
                'END',
                codec)

            with self.assertRaises(asn1tools.errors.Error) as cm:
                module.generate(foo, 'foo')

            self.assertEqual(str(cm.exception),
                             "Foo.A: Unsupported type 'OBJECT IDENTIFIER'.")

    def test_compile_error_unsupported_type_in_sequence(self):
        for codec, module in CODECS_AND_MODULES:
            foo = asn1tools.compile_string(
                'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
                '    A ::= SEQUENCE { '
                '        a NumericString '
                '    } '
                'END',
                codec)

            with self.assertRaises(asn1tools.errors.Error) as cm:
                module.generate(foo, 'foo')

            self.assertEqual(str(cm.exception),
                             "Foo.A.a: Unsupported type 'NumericString'.")

    def test_compile_error_integer_no_minimum_nor_maximum(self):
        for codec, module in CODECS_AND_MODULES:
            foo = asn1tools.compile_string(
                'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
                '    A ::= INTEGER '
                'END',
                codec)

            with self.assertRaises(asn1tools.errors.Error) as cm:
                module.generate(foo, 'foo')

            self.assertEqual(str(cm.exception),
                             "Foo.A: INTEGER has no minimum value.")

    def test_compile_error_integer_no_minimum(self):
        for codec, module in CODECS_AND_MODULES:
            foo = asn1tools.compile_string(
                'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
                '    A ::= INTEGER (MIN..10) '
                'END',
                codec)

            with self.assertRaises(asn1tools.errors.Error) as cm:
                module.generate(foo, 'foo')

            self.assertEqual(str(cm.exception),
                             "Foo.A: INTEGER has no minimum value.")

    def test_compile_error_integer_no_maximum(self):
        for codec, module in CODECS_AND_MODULES:
            foo = asn1tools.compile_string(
                'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
                '    A ::= INTEGER (1..MAX) '
                'END',
                codec)

            with self.assertRaises(asn1tools.errors.Error) as cm:
                module.generate(foo, 'foo')

            self.assertEqual(str(cm.exception),
                             "Foo.A: INTEGER has no maximum value.")

    def test_compile_error_unsigned_integer_over_64_bits(self):
        for codec, module in CODECS_AND_MODULES:
            foo = asn1tools.compile_string(
                'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
                '    A ::= INTEGER (0..18446744073709551616) '
                'END',
                codec)

            with self.assertRaises(asn1tools.errors.Error) as cm:
                module.generate(foo, 'foo')

            self.assertEqual(str(cm.exception),
                             "Foo.A: 18446744073709551616 does not fit in uint64_t.")

    def test_compile_error_unsigned_integer_over_64_signed_bits(self):
        for codec, module in CODECS_AND_MODULES:
            foo = asn1tools.compile_string(
                'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
                '    A ::= INTEGER (-1..9223372036854775808) '
                'END',
                codec)

            with self.assertRaises(asn1tools.errors.Error) as cm:
                module.generate(foo, 'foo')

            self.assertEqual(str(cm.exception),
                             "Foo.A: 9223372036854775808 does not fit in int64_t.")

    def test_compile_error_signed_integer_over_64_bits(self):
        for codec, module in CODECS_AND_MODULES:
            foo = asn1tools.compile_string(
                'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
                '    A ::= INTEGER (-9223372036854775809..0) '
                'END',
                codec)

            with self.assertRaises(asn1tools.errors.Error) as cm:
                module.generate(foo, 'foo')

            self.assertEqual(str(cm.exception),
                             "Foo.A: -9223372036854775809 does not fit in int64_t.")

    def test_compile_error_octet_string_no_size(self):
        for codec, module in CODECS_AND_MODULES:
            foo = asn1tools.compile_string(
                'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
                '    A ::= OCTET STRING '
                'END',
                codec)

            with self.assertRaises(asn1tools.errors.Error) as cm:
                module.generate(foo, 'foo')

            self.assertEqual(str(cm.exception),
                             "Foo.A: OCTET STRING has no maximum length.")

    def test_compile_error_octet_string_no_maximum(self):
        for codec, module in CODECS_AND_MODULES:
            foo = asn1tools.compile_string(
                'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
                '    A ::= OCTET STRING (SIZE(1..MAX)) '
                'END',
                codec)

            with self.assertRaises(asn1tools.errors.Error) as cm:
                module.generate(foo, 'foo')

            self.assertEqual(str(cm.exception),
                             "Foo.A: OCTET STRING has no maximum length.")

    def test_compile_error_sequence_of_no_size(self):
        for codec, module in CODECS_AND_MODULES:
            foo = asn1tools.compile_string(
                'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
                '    A ::= SEQUENCE OF BOOLEAN '
                'END',
                codec)

            with self.assertRaises(asn1tools.errors.Error) as cm:
                module.generate(foo, 'foo')

            self.assertEqual(str(cm.exception),
                             "Foo.A: SEQUENCE OF has no maximum length.")

    def test_compile_error_sequence_of_no_maximum(self):
        for codec, module in CODECS_AND_MODULES:
            foo = asn1tools.compile_string(
                'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
                '    A ::= SEQUENCE (SIZE(1..MAX)) OF BOOLEAN '
                'END',
                codec)

            with self.assertRaises(asn1tools.errors.Error) as cm:
                module.generate(foo, 'foo')

            self.assertEqual(str(cm.exception),
                             "Foo.A: SEQUENCE OF has no maximum length.")

    def test_compile_error_oer_real_not_ieee754(self):
        foo = asn1tools.compile_string(
            'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
            '    A ::= REAL '
            'END',
            'oer')

        with self.assertRaises(asn1tools.errors.Error) as cm:
            asn1tools.source.c.oer.generate(foo, 'foo')

        self.assertEqual(str(cm.exception),
                         "Foo.A: REAL not IEEE 754 binary32 or binary64.")

    def test_compile_error_members_backtrace(self):
        for codec, module in CODECS_AND_MODULES:
            foo = asn1tools.compile_string(
                'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
                '    A ::= SEQUENCE { '
                '        a CHOICE { '
                '            b INTEGER '
                '        } '
                '    } '
                'END',
                codec)

            with self.assertRaises(asn1tools.errors.Error) as cm:
                module.generate(foo, 'foo')

            self.assertEqual(str(cm.exception),
                             "Foo.A.a.b: INTEGER has no minimum value.")

    def test_compile_error_oer_enumerated_min(self):
        foo = asn1tools.compile_string(
            'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
            '    A ::= ENUMERATED { a(-2147483649) } '
            'END',
            'oer')

        with self.assertRaises(asn1tools.errors.Error) as cm:
            asn1tools.source.c.oer.generate(foo, 'foo')

        self.assertEqual(str(cm.exception),
                         "Foo.A: -2147483649 does not fit in int32_t.")

    def test_compile_error_oer_enumerated_max(self):
        foo = asn1tools.compile_string(
            'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
            '    A ::= ENUMERATED { a(2147483649) } '
            'END',
            'oer')

        with self.assertRaises(asn1tools.errors.Error) as cm:
            asn1tools.source.c.oer.generate(foo, 'foo')

        self.assertEqual(str(cm.exception),
                         "Foo.A: 2147483649 does not fit in int32_t.")

    def test_compile_error_bit_strings(self):
        foo = asn1tools.compile_string(
            'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
            '    A ::= BIT STRING (SIZE(1..2))'
            'END',
            'oer')

        with self.assertRaises(asn1tools.errors.Error) as cm:
            asn1tools.source.c.oer.generate(foo, 'foo')

        self.assertEqual(str(cm.exception),
                         "Foo.A: BIT STRING with variable SIZE not supported.")

        foo = asn1tools.compile_string(
            'Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN '
            '    A ::= BIT STRING (SIZE(65))'
            'END',
            'oer')

        with self.assertRaises(asn1tools.errors.Error) as cm:
            asn1tools.source.c.oer.generate(foo, 'foo')

        self.assertEqual(str(cm.exception),
                         "Foo.A: BIT STRING with a length of more than 64 bits are "
                         "not supported.")

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import functools
import asyncio
import websockets
import asn1tools


async def echo(websocket, _, c_source):
    print('Client connected!')

    while True:
        data = await websocket.recv()
        print('Received:', c_source.decode('B', data))
        await websocket.send(data)
        print('Response sent.')


def main():
    c_source = asn1tools.compile_files([
        'asn1/c_source.asn',
        'asn1/programming_types.asn'
    ], 'oer')

    start_server = websockets.serve(
        functools.partial(echo, c_source=c_source),
        'localhost',
        8765)

    print('Listening for websocket clients on localhost:8765...')

    asyncio.get_event_loop().run_until_complete(start_server)
    asyncio.get_event_loop().run_forever()


main()
